@@ -6,6 +6,8 @@ this.SVD_SCREENING = 1e-28
 this.DEFAULT_SYMMETRY = U11
 this.DEFAULT_FLAT = True
 this.DEFAULT_FERMION = True
+this.DEFAULT_LARGE = False
+this.DEFAULT_CUPY = False
 
 symmetry_map = {"U11": U11,
                 "U1": U1,
@@ -17,7 +19,7 @@ def set_symmetry(symmetry):
     if isinstance(symmetry, str):
         symmetry = symmetry.upper()
         if symmetry not in symmetry_map:
-            raise KeyError("input symmetry %s not supported"%symmetry_string)
+            raise KeyError("input symmetry %s not supported" % symmetry)
         this.DEFAULT_SYMMETRY = symmetry_map[symmetry]
     else:
         this.DEFAULT_SYMMETRY = symmetry
@@ -28,22 +30,34 @@ def set_flat(flat):
 def set_fermion(fermion):
     this.DEFAULT_FERMION = fermion
 
+def set_large(large):
+    this.DEFAULT_LARGE = large
+
+def set_cupy(cupy):
+    this.DEFAULT_CUPY = cupy
+
 def set_options(**kwargs):
     symmetry = kwargs.pop("symmetry", None)
     fermion = kwargs.pop("fermion", None)
     flat = kwargs.pop("flat", None)
+    large = kwargs.pop("large", None)
+    cupy = kwargs.pop("cupy", None)
     if symmetry is not None:
         set_symmetry(symmetry)
     if fermion is not None:
         set_fermion(fermion)
     if flat is not None:
         set_flat(flat)
+    if large is not None:
+        set_large(large)
+    if cupy is not None:
+        set_cupy(cupy)
 
 def dispatch_settings(**kwargs):
     keys = list(kwargs.keys())
     for ikey in keys:
         if kwargs.get(ikey) is None:
-            kwargs[ikey] = getattr(this, "DEFAULT_"+ikey.upper())
+            kwargs[ikey] = getattr(this, "DEFAULT_" + ikey.upper())
     _settings = [kwargs.pop(ikey) for ikey in keys]
     if len(_settings) == 1:
         _settings = _settings[0]
