@@ -160,7 +160,7 @@ class MPS(NDArrayOperatorsMixin):
             Delta quantum of MPO operator
     """
 
-    def __init__(self, tensors, const=0, opts=None, dq=None):
+    def __init__(self, tensors, const=0.0, opts=None, dq=None):
         self.tensors = tensors
         self.opts = opts if opts is not None else {}
         self.const = const
