@@ -18,7 +18,7 @@
  *
  */
 
-#include "flat_sparse.hpp"
+#include "flat_fermion.hpp"
 #include <algorithm>
 #include <cstring>
 #include <iostream>
@@ -26,6 +26,84 @@
 #include <math.h>
 #include <numeric>
 
+#define _USE_NEW_PHASE
+
+inline uint8_t compute_phase(const int *perm, int ndim, uint64_t par_pat,
+                             int32_t *rev_idx, int32_t *fperm) {
+    uint64_t tag = 0;
+    int32_t i, k, l;
+    uint8_t n = 0;
+    // get the perm of only fermion ops
+    for (i = 0, k = 0; i < ndim; i++)
+        rev_idx[i] = k, k += (int32_t)((par_pat >> i) & 1);
+    for (i = 0, k = 0; i < ndim; i++)
+        l = perm[i], fperm[k] = rev_idx[l], k += (int32_t)((par_pat >> l) & 1);
+    // n = 1 if number of even cycles in fperm is odd
+    for (i = 0, tag = 0; i < k; i++) {
+        l = i, n ^= !((tag >> l) & 1);
+        while (!((tag >> l) & 1))
+            n ^= 1, tag |= (uint64_t)1 << l, l = fperm[l];
+    }
+    return n;
+}
+
+inline uint8_t compute_phase2(const int *perm, int ndim, int nctr,
+                              uint64_t par_pat, int32_t *rev_idx,
+                              int32_t *fperm) {
+    uint64_t tag = 0;
+    int32_t i, k, l;
+    uint8_t n = 0;
+    // get the perm of only fermion ops
+    for (i = 0, k = 0; i < ndim; i++)
+        rev_idx[i] = k, k += (int32_t)((par_pat >> i) & 1);
+    for (i = 0, k = 0; i < nctr; i++)
+        tag |= (uint64_t)1 << perm[i];
+    for (i = 0, k = 0, tag = ~tag; i < ndim; i++) {
+        fperm[k] = rev_idx[i];
+        k += (int32_t)(((par_pat & tag) >> i) & 1);
+    }
+    for (i = nctr - 1; i >= 0; i--) {
+        l = perm[i];
+        fperm[k] = rev_idx[l];
+        k += (int32_t)((par_pat >> l) & 1);
+    }
+    // n = 1 if number of even cycles in fperm is odd
+    for (i = 0, tag = 0; i < k; i++) {
+        l = i, n ^= !((tag >> l) & 1);
+        while (!((tag >> l) & 1))
+            n ^= 1, tag |= (uint64_t)1 << l, l = fperm[l];
+    }
+    return n;
+}
+
+inline uint8_t compute_phase3(const int *perm, int ndim, int nctr,
+                              uint64_t par_pat, int32_t *rev_idx,
+                              int32_t *fperm) {
+    uint64_t tag = 0;
+    int32_t i, k, l;
+    uint8_t n = 0;
+    // get the perm of only fermion ops
+    for (i = 0, k = 0; i < ndim; i++)
+        rev_idx[i] = k, k += (int32_t)((par_pat >> i) & 1);
+    for (i = 0, k = 0; i < nctr; i++) {
+        l = perm[i];
+        tag |= (uint64_t)1 << l;
+        fperm[k] = rev_idx[l];
+        k += (int32_t)((par_pat >> l) & 1);
+    }
+    for (i = 0, tag = ~tag; i < ndim; i++) {
+        fperm[k] = rev_idx[i];
+        k += (int32_t)(((par_pat & tag) >> i) & 1);
+    }
+    // n = 1 if number of even cycles in fperm is odd
+    for (i = 0, tag = 0; i < k; i++) {
+        l = i, n ^= !((tag >> l) & 1);
+        while (!((tag >> l) & 1))
+            n ^= 1, tag |= (uint64_t)1 << l, l = fperm[l];
+    }
+    return n;
+}
+
 template <typename Q, typename FL>
 void flat_fermion_tensor_transpose(const py::array_t<uint32_t> &aqs,
                                    const py::array_t<uint32_t> &ashs,
@@ -43,7 +121,29 @@ void flat_fermion_tensor_transpose(const py::array_t<uint32_t> &aqs,
     FL *pc = cdata.mutable_data();
 
     const uint32_t *apqs = aqs.data();
-    vector<int> phase_a(n_blocks_a);
+    vector<int8_t> phase_a(n_blocks_a);
+
+#ifdef _USE_NEW_PHASE
+
+    // new impl
+
+    assert(ndima <= 64);
+    vector<int32_t> rev_idx(ndima), fperm(ndima);
+    unordered_map<uint64_t, uint8_t> computed;
+    for (int ia = 0; ia < n_blocks_a; ia++) {
+        uint64_t par_pat = 0;
+        for (int j = 0; j < ndima; j++)
+            par_pat |= (uint64_t)Q::to_q(apqs[ia * asi + j * asj]).parity()
+                       << j;
+        if (!computed.count(par_pat))
+            computed[par_pat] = compute_phase(perma, ndima, par_pat,
+                                              &rev_idx[0], &fperm[0]);
+        phase_a[ia] = 1 - (computed.at(par_pat) << 1);
+    }
+
+#else
+    // old impl
+
     for (int ia = 0; ia < n_blocks_a; ia++) {
         int pnuma[ndima];
         for (int j = 0; j < ndima; j++) {
@@ -66,6 +166,8 @@ void flat_fermion_tensor_transpose(const py::array_t<uint32_t> &aqs,
         phase_a[ia] = pow(-1.0, (double)aparity_counter);
     }
 
+#endif
+
     for (int ia = 0; ia < n_blocks_a; ia++) {
         const FL *a = pa + pia[ia];
         FL *c = pc + pia[ia];
@@ -198,7 +300,7 @@ flat_fermion_tensor_tensordot(
     // contracted q_label hashs
     vector<size_t> ctrqas(n_blocks_a), ctrqbs(n_blocks_b), outqas(n_blocks_a),
         outqbs(n_blocks_b);
-    vector<FL> phase_a(n_blocks_a), phase_b(n_blocks_b);
+    vector<int8_t> phase_a(n_blocks_a), phase_b(n_blocks_b);
 
     psh = aqs.data();
     const uint32_t *apqs = aqs.data();
@@ -206,7 +308,43 @@ flat_fermion_tensor_tensordot(
     for (int i = 0; i < n_blocks_a; i++) {
         ctrqas[i] = q_labels_hash(psh + i * asi, nctr, pidxa, asj);
         outqas[i] = q_labels_hash(psh + i * asi, ndima - nctr, outa, asj);
+    }
 
+    psh = bqs.data();
+    const uint32_t *bpqs = bqs.data();
+    for (int i = 0; i < n_blocks_b; i++) {
+        ctrqbs[i] = q_labels_hash(psh + i * bsi, nctr, pidxb, bsj);
+        outqbs[i] = q_labels_hash(psh + i * bsi, ndimb - nctr, outb, bsj);
+    }
+
+#ifdef _USE_NEW_PHASE
+
+    assert(ndima <= 64 && ndimb <= 64);
+    vector<int32_t> rev_idx(max(ndima, ndimb) + 1), fperm(max(ndima, ndimb) + 1);
+    unordered_map<uint64_t, uint8_t> computed_a, computed_b;
+    for (int i = 0; i < n_blocks_a; i++) {
+        uint64_t par_pat = 0;
+        for (int j = 0; j < ndima; j++)
+            par_pat |= (uint64_t)Q::to_q(apqs[i * asi + j * asj]).parity() << j;
+        if (!computed_a.count(par_pat))
+            computed_a[par_pat] = compute_phase2(ppidxa, ndima, nctr, par_pat,
+                                                &rev_idx[0], &fperm[0]);
+        phase_a[i] = 1 - (computed_a.at(par_pat) << 1);
+    }
+
+    for (int i = 0; i < n_blocks_b; i++) {
+        uint64_t par_pat = 0;
+        for (int j = 0; j < ndimb; j++)
+            par_pat |= (uint64_t)Q::to_q(bpqs[i * bsi + j * bsj]).parity() << j;
+        if (!computed_b.count(par_pat))
+            computed_b[par_pat] = compute_phase3(ppidxb, ndimb, nctr, par_pat,
+                                                &rev_idx[0], &fperm[0]);
+        phase_b[i] = 1 - (computed_b.at(par_pat) << 1);
+    }
+
+#else
+
+    for (int i = 0; i < n_blocks_a; i++) {
         int pnuma[ndima];
         for (int j = 0; j < ndima; j++) {
             pnuma[j] = Q::to_q(apqs[i * asi + j * asj]).parity();
@@ -228,12 +366,7 @@ flat_fermion_tensor_tensordot(
         phase_a[i] = pow(-1.0, (double)aparity_counter);
     }
 
-    psh = bqs.data();
-    const uint32_t *bpqs = bqs.data();
     for (int i = 0; i < n_blocks_b; i++) {
-        ctrqbs[i] = q_labels_hash(psh + i * bsi, nctr, pidxb, bsj);
-        outqbs[i] = q_labels_hash(psh + i * bsi, ndimb - nctr, outb, bsj);
-
         int pnumb[ndimb];
         for (int j = 0; j < ndimb; j++) {
             pnumb[j] = Q::to_q(bpqs[i * bsi + j * bsj]).parity();
@@ -255,6 +388,8 @@ flat_fermion_tensor_tensordot(
         phase_b[i] = pow(-1.0, (double)bparity_counter);
     }
 
+#endif
+
     unordered_map<size_t, vector<int>> map_idx_b;
     for (int i = 0; i < n_blocks_b; i++)
         map_idx_b[ctrqbs[i]].push_back(i);
@@ -396,7 +531,7 @@ flat_fermion_tensor_tensordot(
             int xia = 0, xib = 0;
             for (size_t i = 0; i < mmq.second.size(); i++) {
                 xia = mmq.second[i].first, xib = mmq.second[i].second;
-                FL phase = phase_a[xia] * phase_b[xib];
+                FL phase = (FL)phase_a[xia] * (FL)phase_b[xib];
                 int ldb = trans_b == 1 ? b_free_dim[xib] : ctr_dim[xia];
                 int lda = trans_a == -1 ? ctr_dim[xia] : a_free_dim[xia];
                 int ldc = b_free_dim[xib];
@@ -417,52 +552,6 @@ flat_fermion_tensor_tensordot(
     return std::make_tuple(cqs, cshs, cdata, cidxs);
 }
 
-template <typename Q>
-tuple<py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<uint64_t>>
-flat_fermion_tensor_skeleton(const vector<map_uint_uint<Q>> &infos,
-                             uint32_t fdq) {
-    int ndim = (int)infos.size();
-    size_t nx = 1;
-    for (int i = 0; i < ndim; i++) {
-        nx *= infos[i].size();
-    }
-    const string pattern(ndim, '+');
-    vector<uint32_t> qs, shs;
-    vector<uint64_t> idxs(1, 0);
-    vector<vector<pair<Q, uint32_t>>> infox;
-    bond_info_trans<Q>(infos, pattern, infox, true);
-    Q dq = Q::to_q(fdq);
-    vector<uint32_t> qk(ndim), shk(ndim);
-    for (size_t x = 0; x < nx; x++) {
-        size_t xp = x;
-        Q xq;
-        for (int i = ndim - 1; i >= 0; xp /= infox[i].size(), i--)
-            xq = xq + infox[i][xp % infox[i].size()].first;
-
-        if (xq.parity() == dq.parity()) {
-            uint32_t sz = 1;
-            xp = x;
-            for (int i = ndim - 1; i >= 0; xp /= infox[i].size(), i--) {
-                auto &r = infox[i][xp % infox[i].size()];
-                qk[i] = Q::from_q(r.first);
-                shk[i] = r.second, sz *= r.second;
-            }
-            qs.insert(qs.end(), qk.begin(), qk.end());
-            shs.insert(shs.end(), shk.begin(), shk.end());
-            idxs.push_back(idxs.back() + sz);
-        }
-    }
-    vector<ssize_t> sh = {(ssize_t)qs.size() / ndim, ndim};
-    py::array_t<uint32_t> cqs(sh), cshs(sh),
-        cidxs(vector<ssize_t>{(ssize_t)idxs.size()});
-    assert(cqs.strides()[1] == sizeof(uint32_t));
-    assert(cshs.strides()[1] == sizeof(uint32_t));
-    memcpy(cqs.mutable_data(), qs.data(), qs.size() * sizeof(uint32_t));
-    memcpy(cshs.mutable_data(), shs.data(), shs.size() * sizeof(uint32_t));
-    memcpy(cidxs.mutable_data(), idxs.data(), idxs.size() * sizeof(uint64_t));
-    return std::make_tuple(cqs, cshs, cidxs);
-}
-
 template <typename Q>
 map_fusing flat_fermion_tensor_kron_sum_info(const py::array_t<uint32_t> &aqs,
                                              const py::array_t<uint32_t> &ashs,
@@ -676,8 +765,8 @@ flat_fermion_tensor_qr(const py::array_t<uint32_t> &aqs,
     FL *pl = ldata.mutable_data(), *pr = rdata.mutable_data();
 
     int lwork = (is_qr ? max_rshape : max_lshape) * 34, info = 0;
-    vector<FL> tau(max_mshape), work(lwork), tmpl(max_tmp_size),
-        tmpr(max_tmp_size);
+    vector<FL> tau(max_mshape), work, tmpl(max_tmp_size), tmpr(max_tmp_size);
+    work.reserve(lwork);
     int iis = 0, iil = 0, iir = 0;
     plidxs[0] = pridxs[0] = 0;
     if (is_qr)
@@ -689,9 +778,13 @@ flat_fermion_tensor_qr(const py::array_t<uint32_t> &aqs,
         FL *mat = mat_data.data() + mq.second;
         int ml = linfo.at(ql).first, mr = rinfo.at(qr).first, mm = min(ml, mr);
         const int lshape = ml, rshape = mr, mshape = mm;
-        int lwork = max(ml, mr) * 34;
         if (is_qr) {
             memcpy(tmpr.data(), mat, sizeof(FL) * lshape * rshape);
+            int qlwork = -1;
+            xgelqf<FL>(&rshape, &lshape, tmpr.data(), &rshape, tau.data(),
+                       work.data(), &qlwork, &info);
+            if (lwork < (int)abs(work[0]))
+                lwork = (int)abs(work[0]), work.reserve(lwork);
             xgelqf<FL>(&rshape, &lshape, tmpr.data(), &rshape, tau.data(),
                        work.data(), &lwork, &info);
             assert(info == 0);
@@ -701,6 +794,11 @@ flat_fermion_tensor_qr(const py::array_t<uint32_t> &aqs,
             assert(info == 0);
         } else {
             memcpy(tmpl.data(), mat, sizeof(FL) * lshape * rshape);
+            int qlwork = -1;
+            xgeqrf<FL>(&rshape, &lshape, tmpl.data(), &rshape, tau.data(),
+                       work.data(), &qlwork, &info);
+            if (lwork < (int)abs(work[0]))
+                lwork = (int)abs(work[0]), work.reserve(lwork);
             xgeqrf<FL>(&rshape, &lshape, tmpl.data(), &rshape, tau.data(),
                        work.data(), &lwork, &info);
             assert(info == 0);
@@ -771,9 +869,15 @@ flat_fermion_tensor_qr(const py::array_t<uint32_t> &aqs,
 #define TMPL_NAME flat_fermion
 
 #include "symmetry_tmpl.hpp"
+
+#define TMPL_FL float
+#include "symmetry_tmpl.hpp"
+#undef TMPL_FL
+
 #define TMPL_FL double
 #include "symmetry_tmpl.hpp"
 #undef TMPL_FL
+
 #define TMPL_FL complex<double>
 #include "symmetry_tmpl.hpp"
 #undef TMPL_FL
