@@ -106,7 +106,7 @@ flat_sparse_right_canonicalize_indexed<TMPL_Q, TMPL_FL>(
 TMPL_EXTERN template tuple<
     py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
     py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-    py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+    py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
     py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>>
 flat_sparse_left_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
                                       const py::array_t<uint32_t> &ashs,
@@ -116,7 +116,7 @@ flat_sparse_left_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
 TMPL_EXTERN template tuple<
     py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
     py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-    py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+    py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
     py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>>
 flat_sparse_right_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
                                        const py::array_t<uint32_t> &ashs,
@@ -126,7 +126,7 @@ flat_sparse_right_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
 TMPL_EXTERN template pair<
     tuple<py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
           py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-          py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+          py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
           py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>>,
     py::array_t<uint32_t>>
 flat_sparse_left_svd_indexed<TMPL_Q, TMPL_FL>(
@@ -136,7 +136,7 @@ flat_sparse_left_svd_indexed<TMPL_Q, TMPL_FL>(
 TMPL_EXTERN template pair<
     tuple<py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
           py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-          py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+          py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
           py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>>,
     py::array_t<uint32_t>>
 flat_sparse_right_svd_indexed<TMPL_Q, TMPL_FL>(
@@ -146,7 +146,7 @@ flat_sparse_right_svd_indexed<TMPL_Q, TMPL_FL>(
 TMPL_EXTERN template tuple<
     py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
     py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-    py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+    py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
     py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>>
 flat_sparse_tensor_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
                                         const py::array_t<uint32_t> &ashs,
@@ -159,16 +159,16 @@ flat_sparse_tensor_svd<TMPL_Q, TMPL_FL>(const py::array_t<uint32_t> &aqs,
 TMPL_EXTERN template tuple<
     py::array_t<uint32_t>, py::array_t<uint32_t>, py::array_t<TMPL_FL>,
     py::array_t<uint64_t>, py::array_t<uint32_t>, py::array_t<uint32_t>,
-    py::array_t<double>, py::array_t<uint64_t>, py::array_t<uint32_t>,
-    py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>, double>
+    py::array_t<TMPL_FP>, py::array_t<uint64_t>, py::array_t<uint32_t>,
+    py::array_t<uint32_t>, py::array_t<TMPL_FL>, py::array_t<uint64_t>, TMPL_FP>
 flat_sparse_truncate_svd<TMPL_Q, TMPL_FL>(
     const py::array_t<uint32_t> &lqs, const py::array_t<uint32_t> &lshs,
     const py::array_t<TMPL_FL> &ldata, const py::array_t<uint64_t> &lidxs,
     const py::array_t<uint32_t> &sqs, const py::array_t<uint32_t> &sshs,
-    const py::array_t<double> &sdata, const py::array_t<uint64_t> &sidxs,
+    const py::array_t<TMPL_FP> &sdata, const py::array_t<uint64_t> &sidxs,
     const py::array_t<uint32_t> &rqs, const py::array_t<uint32_t> &rshs,
     const py::array_t<TMPL_FL> &rdata, const py::array_t<uint64_t> &ridxs,
-    int max_bond_dim, double cutoff, double max_dw, double norm_cutoff,
+    int max_bond_dim, TMPL_FP cutoff, TMPL_FP max_dw, TMPL_FP norm_cutoff,
     bool eigen_values);
 
 #else
